;;;; regex-chardefs.scm
;;;; Kon Lovett, Feb '08

(use regex)

(declare
  (usual-integrations)
  (disable-interrupts)
  (generic) ; PCRE options use lotsa bits
  #;(disable-warning var)
  (no-bound-checks)
  (no-procedure-checks-for-usual-bindings)
  (export
    regex-chardef-set!
    regex-chardefs-update!
    regex-chardefs ) )

(cond-expand
 [unsafe
  (eval-when (compile)
    (define-macro (##sys#check-chardef-table . _) '(##core#undefined))
    (define-macro (##sys#check-integer . _) '(##core#undefined))
    (define-macro (##sys#check-blob . _) '(##core#undefined))
    (define-macro (##sys#check-vector . _) '(##core#undefined))
    (define-macro (##sys#check-structure . _) '(##core#undefined))
    (define-macro (##sys#check-range . _) '(##core#undefined))
    (define-macro (##sys#check-pair . _) '(##core#undefined))
    (define-macro (##sys#check-list . _) '(##core#undefined))
    (define-macro (##sys#check-symbol . _) '(##core#undefined))
    (define-macro (##sys#check-string . _) '(##core#undefined))
    (define-macro (##sys#check-char . _) '(##core#undefined))
    (define-macro (##sys#check-exact . _) '(##core#undefined))
    (define-macro (##sys#check-port . _) '(##core#undefined))
    (define-macro (##sys#check-number . _) '(##core#undefined))
    (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
 [else] )

;;;

#>
#include "config.h"
#include "pcre.h"
#include "pcre_internal.h"
<#

;;; Character Definition Tables:

;; Character Class

(define-foreign-enum (pcre-cbit unsigned-int)
  (space cbit_space)
  (xdigit cbit_xdigit)
  (digit cbit_digit)
  (upper cbit_upper)
  (lower cbit_lower)
  (word cbit_word)
  (graph cbit_graph)
  (print cbit_print)
  (punct cbit_punct)
  (cntrl cbit_cntrl) )

(define cbit-symbols '(space xdigit digit upper lower word graph print punct cntrl))

;; Character Type

(define-foreign-enum (pcre-ctype unsigned-int)
  (space ctype_space)
  (letter ctype_letter)
  (digit ctype_digit)
  (xdigit ctype_xdigit)
  (word ctype_word)
  (meta ctype_meta) )

(define ctype-symbols '(space xdigit digit letter word meta))

;; Accessors

(define chardef-lower-case-set!
  (foreign-lambda* void (((c-pointer unsigned-char) tables) (int idx) (unsigned-char lower))
    "tables[lcc_offset + idx] = lower;"))

(define chardef-lower-case
  (foreign-lambda* unsigned-int (((const (c-pointer unsigned-char)) tables) (int idx))
    "return(tables[lcc_offset + idx]);"))

(define chardef-flipped-case-set!
  (foreign-lambda* void (((c-pointer unsigned-char) tables) (int idx) (unsigned-char flipped))
    "tables[fcc_offset + idx] = flipped;"))

(define chardef-flipped-case
  (foreign-lambda* unsigned-int (((const (c-pointer unsigned-char)) tables) (int idx))
    "return (tables[fcc_offset + idx]);"))

(define chardef-class-clear!
  (foreign-lambda* void (((c-pointer unsigned-char) tables) (int idx) (unsigned-int class))
    "tables[cbits_offset + class + idx/8] &= ~(1 << (idx & 7));"))

(define chardef-class-set!
  (foreign-lambda* void (((c-pointer unsigned-char) tables) (int idx) (unsigned-int class))
    "tables[cbits_offset + class + idx/8] |= 1 << (idx & 7);"))

(define chardef-class
  (foreign-lambda* unsigned-int (((const (c-pointer unsigned-char)) tables) (int idx) (unsigned-int class))
    "return (tables[cbits_offset + class + idx/8] & (1 << (idx & 7)));"))

(define (chardef-classes-clear! tables idx)
  (for-each
    (lambda (sym)
      (chardef-class-clear! tables idx (pcre-cbit->number sym)))
    cbit-symbols) )

(define chardef-type-clear!
  (foreign-lambda* void (((c-pointer unsigned-char) tables) (int idx))
    "tables[ctypes_offset + idx] = 0;"))

(define chardef-type-set!
  (foreign-lambda* void (((c-pointer unsigned-char) tables) (int idx) (unsigned-int type))
    "tables[ctypes_offset + idx] += type;"))

(define chardef-type
  (foreign-lambda* unsigned-int (((const (c-pointer unsigned-char)) tables) (int idx) (unsigned-int type))
    "return (tables[ctypes_offset + idx] & type);"))

;; Update a character definition.
;;
;; 'tables' is a character definition tables structure
;; 'chardef' is a length 4 vector where
;; element 0 is the lower-case character or #f
;; element 1 is the flipped-case character or #f
;; element 2 is a list of character class names or #f
;; element 3 is a list of character type names or #f

(define (regex-chardef-set! tables idx-or-char chardef)
  ; Check proper character definition tables structure
  (##sys#check-chardef-table tables 'regex-chardef-set!)
  ; Need character table index
  (let ([idx (if (char? idx-or-char) (char->integer idx-or-char) idx-or-char)])
    ; Check proper index
    (##sys#check-exact idx 'regex-chardef-set!)
    (unless (and (fx<= 0 idx) (fx<= idx 255))
      (##sys#error 'regex-chardef-set! "invalid character index - must be in [0 255]" idx) )
    ; Check proper character definition structure
    (##sys#check-vector chardef 'regex-chardef-set!)
    (unless (fx= 4 (vector-length chardef))
      (##sys#error 'regex-chardef-set! "invalid chardef length - must be 4" chardef) )
    ; Change lower case character?
    (and-let* ([lower (vector-ref chardef 0)])
      (chardef-lower-case-set! tables idx lower) )
    ; Change flipped case character?
    (and-let* ([flipped (vector-ref chardef 1)])
       (chardef-flipped-case-set! tables idx flipped) )
    ;
    (let ([set-symbols
            (lambda (syms clear sym->num set)
              (##sys#check-list syms 'regex-chardef-set!)
              (clear tables idx)
              (for-each
                (lambda (sym)
                  (set tables idx (sym->num sym)))
                syms) )])
      ; Change character class?
      (and-let* ([cbits (vector-ref chardef 2)])
        (set-symbols cbits chardef-classes-clear! pcre-cbit->number chardef-class-set!) )
      ; Change character type?
      (and-let* ([ctypes (vector-ref chardef 3)])
        (set-symbols ctypes chardef-type-clear! pcre-ctype->number chardef-type-set!) ) )
    (void) ) )

;; Update character definition tables.
;;
;; 'tables' is a character definition tables structure
;; 'chardefs' is a length 256 vector of #f or character definitions

(define (regex-chardefs-update! tables chardefs)
  ; Check proper character definition tables structure
  (##sys#check-chardef-table tables 'regex-chardef-update!)
  ; Check proper character definition table structure
  (##sys#check-vector chardefs 'regex-chardefs-update!)
  (unless (fx= 256 (vector-length chardefs))
    (##sys#error 'regex-chardefs-update! "invalid chardefs length - must be 256" chardefs) )
  ; Set every character definition
  (do ([idx 0 (fx+ idx 1)])
      [(fx= 256 idx)]
    ; When a new character definition
    (and-let* ([chardef (vector-ref chardefs idx)])
      (regex-chardef-set! tables idx chardef) ) ) )

;; Get the character definitions.
;; Returns a character definitions vector.

(define (regex-chardefs tables)
  ; Check proper character definition tables structure
  (##sys#check-chardef-table tables 'regex-chardefs)
  ; 256 character definitions
  (let ([chardefs (make-vector 256)])
    ; Get every character definition
    (do ([idx 0 (fx+ idx 1)])
        [(fx= 256 idx) chardefs]
      ; This character definition
      (let ([chardef (make-vector 4)])
        (vector-set! chardefs idx chardef)
        ; Lower-case
        (vector-set! chardef 0 (integer->char (chardef-lower-case tables idx)))
        ; Flipped-case
        (vector-set! chardef 1 (integer->char (chardef-flipped-case tables idx)))
        ;
        (let ([get-symbols
                (lambda (syms get sym->num)
                  (let loop ([syms syms]
                             [lst '()])
                    (if (null? syms)
                        lst
                        (let ([sym (car syms)])
                          (loop (cdr syms)
                                (if (fx= 0 (get tables idx (sym->num sym)))
                                    lst
                                    (cons sym lst) ) ) ) ) ) )])
          ; Character class
          (vector-set! chardef 2 (get-symbols cbit-symbols chardef-class pcre-cbit->number))
          ; Character type
          (vector-set! chardef 3 (get-symbols ctype-symbols chardef-type pcre-ctype->number)) ) ) ) ) )
